@@ -40,6 +40,7 @@ namespace {
 MappingOptions::MappingOptions(pTHX_ SV *options_ref) :
         use_bigints(sizeof(IV) < sizeof(int64_t)),
         check_required_fields(true),
+        ignore_undef_fields(false),
         explicit_defaults(false),
         encode_defaults(false),
         encode_defaults_proto3(false),
@@ -77,6 +78,7 @@ MappingOptions::MappingOptions(pTHX_ SV *options_ref) :
     BOOLEAN_OPTION(implicit_maps, implicit_maps);
     BOOLEAN_OPTION(decode_blessed, decode_blessed);
     BOOLEAN_OPTION(fail_ref_coercion, fail_ref_coercion);
+    BOOLEAN_OPTION(ignore_undef_fields, ignore_undef_fields);
 
 #undef BOOLEAN_OPTION
 
@@ -524,6 +526,7 @@ void Dynamic::bind_message(pTHX_ const string &perl_package, Mapper *mapper, con
     }
 
     copy_and_bind(aTHX_ "set_decoder_options", perl_package, mapper);
+    copy_and_bind(aTHX_ "set_encoder_options", perl_package, mapper);
     copy_and_bind(aTHX_ "decode_upb", perl_package, mapper);
     copy_and_bind(aTHX_ "decode_bbpb", perl_package, mapper);
     if (options.default_decoder == MappingOptions::Upb) {
